from pygame.sprite import Sprite, Group
from pygame import Surface, Rect, transform
from util import Timer
import math
import data
import logging
import random

class Multisprite(Sprite):
  """
  Image contains multiple sprites, and one of these
  is active at a time
  """
  def __init__(self, pos, image, groups=(), size=(1, 1), frames=((0, 0),)):
    Sprite.__init__(self, *groups)
    width, height = image.get_size()
    frame_height = height/ size[0]
    frame_width = width / size[1]
    self.surfaces = []
    for i in xrange(size[0]):
      row = []
      for j in xrange(size[1]):
        rect = Rect(j * frame_width, i * frame_height, frame_width, frame_height)
        row.append(image.subsurface(rect))
      self.surfaces.append(row)
    self.image = self.surfaces[0][0]
    self.rect = self.image.get_rect()
    if len(pos) > 2: # if pos contains multiple points
      self.rect.topleft = pos[0]
    else:
      self.rect.midtop = pos
    self.frames = frames
    self.current_frame = 0

  def set_sprite(self, row, col):
    self.image = self.surfaces[row][col]

  def animate(self):
    num_frames = len(self.frames)
    self.current_frame = (self.current_frame + 1) % num_frames
    self.set_sprite(*self.frames[self.current_frame])

class MovingThing(Multisprite):
  def __init__(self, pos, image, config, speed=0, health=100, **kwargs):
    Multisprite.__init__(self, pos, image, **kwargs)
    self.colliderect = self.rect.inflate(-10, -5)
    self.speed = speed
    self.health = health
    self.dead = False

  def moveright(self, amount):
    if self.dead:
      return
    self.rect.left += amount * self.speed
    self.colliderect.center = self.rect.center

  def movedown(self, amount):
    if self.dead:
      return
    self.rect.top += amount * self.speed
    self.colliderect.center = self.rect.center

class Explosion(Multisprite):
  def __init__(self, pos):
    surface = data.image_load('explosion.png')
    Multisprite.__init__(self, pos, surface, size=(1, 6), \
        frames = [(0, i) for i in range(6)])
    self.finished = False

  def animate(self):
    num_frames = len(self.frames)
    self.current_frame = (self.current_frame + 1) #% num_frames
    # remove explosion after animation ends
    if self.current_frame > num_frames-1:
      self.finished = True
    else: # explosion didn't end, update frame
      self.set_sprite(*self.frames[self.current_frame]) 

class PlayerSprite(MovingThing):
  def __init__(self, pos, config):
    surface = data.image_load('diver.png')
    MovingThing.__init__(self, pos, surface, config, size=(1, 10), \
        speed = config.getfloat('player', 'speed'), \
        frames = [(0, i) for i in range(10)] \
    )
    self.blank = Surface(self.image.get_size())
    self.blank.set_colorkey((0, 0, 0))
    self.bullet_speed = config.getfloat('player', 'bullet_speed')
    self.stunned = False
    self.stun_timer = Timer(config.getint('game', 'stun_time'),
        self.unstun)
    self.respawn_timer = Timer(500, self.respawn)
    self.lives = 5
    self.initial = pos
    self.invulnerable = False
    self.bullet_upgrades = []

  def die(self, explosion_groups):
    if self.dead:
      return
    self.dead = True
    self.stunned = False
    self.lives -= 1
    self.respawn_timer.reset()
    pos = (self.rect.center[0], self.rect.top) # x=center of rect, y=top of rect
    explosion = Explosion(pos)
    for group in explosion_groups:
      group.add(explosion)
    self.rect.midtop = self.initial

  def respawn(self):
    self.dead = False
    self.stun() # start off stunned
    self.invulnerable = True

  def shoot(self, *args):
    logging.info('KAPOW')
    bullet = PlayerBullet(self.rect.midbottom, self.bullet_speed)
    if self.bullet_upgrades:
      bullet.change_stats(self.bullet_upgrades[0], self.bullet_upgrades[1], self.bullet_upgrades[2])
    return bullet

  def update(self, ms, depth):
    if self.stunned:
      self.stun_timer.update(ms)
    elif self.dead:
      self.respawn_timer.update(ms)
    # WEAPON UPGRADES!
    if depth >= 15000:
      self.bullet_upgrades = [(4,4), (255,255,255), 100]
    #! INSERT MORE DEPTHS AND UPGRADES?
    

  def animate(self):
    if self.stunned:
      if self.image != self.blank:
        self.image = self.blank
      else:
        self.set_sprite(*self.frames[self.current_frame])
    else:
      MovingThing.animate(self)

  def stun(self):
    if not self.stunned and not self.dead:
      self.stunned = True
      self.stun_timer.reset()
      logging.info('OUCH')

  def unstun(self):
    self.stunned = False
    self.invulnerable = False

class Enemy(MovingThing):
  # Firing strategies
  UNARMED = 0
  SHOOT_FWD = 1
  AIM = 2

  def __init__(self, pos, config, shoot=0, bullet_type=None, height=700):
    self.shoot = shoot
    self.loop_pos = 0 # current index for points in loop
    self.speed = [0,0]
    self.speed_mult = 0.2 # speed multiplier
    self.bullet_type = bullet_type
    self.moving = True
    self.loop_path = False
    self.fire_multi = False
    self.config = config
    self.height = height
    if isinstance(pos, tuple):
      pos = [pos]
    self.points = pos # for enemies in loop pattern
    if shoot != 0: # 1 = regular, 2 = aimed at player
      self.bullet_speed = 0.3
      self.firing_timer = 0
      self.fire_wait_time = 2000

  @property
  def has_path(self):
    return len(self.points) > 1

  def update(self, ms, player, mixer, enemy_bullets=(), enemy_groups=()):
    self.move_path(ms, player, mixer, enemy_bullets)
    self.fire(ms, player, mixer, self.rect, enemy_bullets, enemy_groups, self.config)

  def move_path(self, ms, player, mixer, enemy_bullets=[]):
    # Move along the path
    if self.has_path:
      # CHECK IF ON/PAST A POINT
      # (The +/- 2 is to compensate for when coords are +/- 2 off)
      if int(self.rect.x) >= self.points[self.loop_pos][0]-2 and int(self.rect.x) <= self.points[self.loop_pos][0]+2 \
      and int(self.rect.y) >= self.points[self.loop_pos][1]-2 and int(self.rect.y) <= self.points[self.loop_pos][1]+2 \
      and self.moving:
        if self.loop_path:
          self.loop_pos = (self.loop_pos + 1) % len(self.points)
        elif self.loop_pos < len(self.points) - 1:
          self.loop_pos += 1

      overshot = False
      if self.loop_pos < len(self.points):
        # MOVE TOWARDS NEXT POINT (CALC SPEEDS)
        diff_x = self.points[self.loop_pos][0] - self.rect.x
        diff_y = self.points[self.loop_pos][1] - self.rect.y
        theta = math.atan2(diff_y, diff_x)
        self.speed = self.speed_mult*math.cos(theta), self.speed_mult*math.sin(theta)
        distance_sq = diff_x ** 2 + diff_y ** 2
        overshot = (self.speed_mult * ms) ** 2 > distance_sq
      if overshot:
        # Stop at the next point to fire
        self.rect.x = self.points[self.loop_pos][0]
        self.rect.y = self.points[self.loop_pos][1]
      else:
        # Keep going
        self.rect.x += self.speed[0] * ms
        self.rect.y += self.speed[1] * ms
      self.colliderect.center = self.rect.center


  def fire(self, ms, player, mixer, rect, enemy_bullets, enemy_groups, config):
    # SHOOTING
    if self.bullet_type == None:
      return
    if self.shoot != Enemy.UNARMED:
      if rect.top >= self.height or rect.bottom < 0:
        return
      self.firing_timer -= ms
      if self.firing_timer > 0:
        return
      mixer.play_sound('inkball')
      logging.info("SQUIPOW")
      yspeed = -self.bullet_speed # it's a feature
      if self.shoot == Enemy.AIM:
        dx, dy = (player.rect.x - rect.x), (player.rect.y - rect.y) # difference between player and squid
        try:
          xspeed = dx * (self.bullet_speed / dy)
        except ZeroDivisionError:
          xspeed = 0
        if xspeed > 0.3:
          xspeed = 0.3
        elif xspeed < -0.3:
          xspeed = -0.3
      else:
        if hasattr(self, 'rotation'):
          xspeed = - self.bullet_speed * math.cos(self.rotation)
          yspeed = self.bullet_speed * math.sin(self.rotation)

          xspeed2 = - self.bullet_speed * math.cos(self.rotation - math.pi / 4)
          yspeed2 = self.bullet_speed * math.sin(self.rotation - math.pi / 4)

          xspeed3 = - self.bullet_speed * math.cos(self.rotation + math.pi / 4)
          yspeed3 = self.bullet_speed * math.sin(self.rotation + math.pi / 4)
        else:
          xspeed = 0
      self.firing_timer = self.fire_wait_time
      if self.bullet_type != None:
        # Midtop is a stupid fucking coordinate what the fuck was I thinking fuck fuck fuck
        if hasattr(self, 'rotation'):
          midtop = (rect.centerx - 2, rect.centery - 2) # Everything is about 4px. Do not question.
        else:
          midtop = rect.midtop
        if self.bullet_type is Jellyfish:
          target = (random.randint(0, 20) + midtop[0], random.randint(0, 20) + midtop[1])
          sprite = Jellyfish(midtop, config, target=target)
          for group in enemy_groups:
            group.add(sprite)
        elif self.fire_multi and hasattr(self, 'rotation'):
          bullet = self.bullet_type(midtop, (yspeed, -xspeed))
          enemy_bullets.add(bullet)
          bullet = self.bullet_type(midtop, (yspeed2, -xspeed2))
          enemy_bullets.add(bullet)
          bullet = self.bullet_type(midtop, (yspeed3, -xspeed3))
          enemy_bullets.add(bullet)
        else:
          bullet = self.bullet_type(midtop, (yspeed, -xspeed))
          enemy_bullets.add(bullet)


# TODO these should move up a bit faster than the background, to match the animation
class Jellyfish(Enemy):
  def __init__(self, pos, config, shoot=0, target=None):
    if target:
      path = [pos, target]
    else:
      path = pos
    Enemy.__init__(self, path, config, shoot)
    surface = data.image_load('jellyfish.png')
    MovingThing.__init__(self, pos, surface, config, size=(1, 8), \
        frames=[(0, i) for i in range(8)]
    )
    self.colliderect = self.rect.copy()
    self.loop_path = False

class Urchin(Enemy):
  def __init__(self, pos, config, shoot=0):
    Enemy.__init__(self, pos, config, shoot)
    surface = data.image_load('urchin.png')
    MovingThing.__init__(self, pos, surface, config, size=(1, 3), \
        frames=[(0, i) for i in range(3)]
    )
    self.colliderect = self.rect.copy()

class Squid(Enemy):
  def __init__(self, pos, config, shoot=0):
    Enemy.__init__(self, pos, config, shoot, SquidBullet)
    surface = data.image_load('squid.png')
    MovingThing.__init__(self, pos, surface, config, size=(1, 6), \
        frames=[(0, i) for i in range(6)]
    )
    self.colliderect = self.rect.copy()

class Squid2(Enemy):
  def __init__(self, pos, config, shoot=0):
    Enemy.__init__(self, pos, config, shoot, Squid2Bullet)
    surface = data.image_load('squid.png')
    MovingThing.__init__(self, pos, surface, config, size=(1, 6), \
        frames=[(0, i) for i in range(6)]
    )
    self.colliderect = self.rect.copy()
    self.speed_mult = 0.3
    if self.shoot != 0:
      self.bullet_speed = 0.15
      self.fire_wait_time = 750

class Arm(Enemy):
  # TODO need animations for this, including hit animations
  # and different guns for each arm
  def __init__(self, pos, surface, config, shoot=0, bullet_speed=0.15, bullet_type=None, fire_wait_time=1000, fire_multi=False, **kwargs):
    Enemy.__init__(self, pos, config, shoot, bullet_type)
    MovingThing.__init__(self, pos, surface, config, **kwargs)
    self.colliderect = self.rect.copy()
    self.speed_mult = 0.3
    if self.shoot != 0:
      self.bullet_speed = bullet_speed
      self.fire_wait_time = fire_wait_time
      self.fire_multi = fire_multi

class Nemesis(Enemy):
  '''Nemesis is made up of smaller parts'''
  class NOOOOOooo(Exception):
    pass
  def __init__(self, pos, config, width, height):
    Enemy.__init__(self, pos, config)
    self.size = 250
    self.radius = 40
    self.spacing = 50
    self.config = config
    total = (self.radius + self.spacing) * 2
    surface = Surface((total, total))
    MovingThing.__init__(self, pos, surface, config)
    self.transparent = (255, 255, 0)
    self.image.set_colorkey(self.transparent)
    self.arms = Group()
    self.cyborg_arms = Group()
    self.rotation = 0
    self.speed_mult = 0.25
    self.dead = False
    self.positions = {}
    self.body = data.image_load('nemesis_body.png')
    self.cannon = data.image_load('nemesis_cannon.png')
    self.gatling = data.image_load('nemesis_gatling.png')
    self.jar = data.image_load('nemesis_jar.png')
    self.laser = data.image_load('nemesis_laser.png')
    self.multi = data.image_load('nemesis_multi.png')
    self.tentacle = data.image_load('nemesis_tentacle.png')

    self.head = Multisprite((0, 0), self.body, size=(1, 2), frames=((0, 0), (0, 1)))
    self.dead = False
    self.toggle = 0

    self.move_max_time = 5000
    self.action_timer = self.move_max_time
    self.fire_max_time = 5000
    self.current_arm = None
    self.firing = False
    self.loop_path = True
    self.health = 5000

    healths = range(1000, 10000, 1000)
    #healths = []
    random.shuffle(healths)

    for i, health in zip(xrange(8), healths):
      if i == 1 or i == 3:
        arm = Arm((0, 0), self.tentacle.copy(), config, health=health, shoot=Enemy.UNARMED,
          size=(1, 6), frames=[(0, j) for j in range(6)])
      elif i == 2:
        arm = Arm((0, 0), self.laser.copy(), config, health=health, shoot=Enemy.SHOOT_FWD,
          size=(1, 2), frames=[(0, j) for j in range(2)], bullet_type=NotPlayerBullet, bullet_speed=0.4, fire_wait_time=300)
      elif i == 4 or i == 0:
        arm = Arm((0, 0), self.gatling.copy(), config, health=health, shoot=Enemy.SHOOT_FWD,
          size=(1, 2), frames=[(0, j) for j in range(2)], bullet_type=SquidBullet, bullet_speed=0.2, fire_wait_time = 100)
      elif i == 5:
        arm = Arm((0, 0), self.jar.copy(), config, health=health, shoot=Enemy.SHOOT_FWD,
          size=(1, 2), frames=[(0, j) for j in range(2)], bullet_type=Jellyfish, fire_wait_time = 500)
        self.jellyfish_arm = arm
      elif i == 6:
        arm = Arm((0, 0), self.multi.copy(), config, health=health, shoot=Enemy.SHOOT_FWD,
          size=(1, 2), frames=[(0, j) for j in range(2)], bullet_type=SquidBullet, bullet_speed=0.3, fire_wait_time = 250, fire_multi=True)
      elif i == 7:
        arm = Arm((0, 0), self.cannon.copy(), config, health=health, shoot=Enemy.SHOOT_FWD,
            bullet_type=Squid2Bullet, bullet_speed=0.6, fire_wait_time=800)

      self.arm_image = arm.image
      self.arms.add(arm)
      if i not in (1, 3):
        self.cyborg_arms.add(arm)
      self.positions[arm] = i

    # Paths. Points correspond to nemesis topleft.
    left = - self.spacing / 2
    right = width - self.rect.width + self.spacing / 2
    top =  - self.spacing / 2
    bottom = height - self.rect.height + self.spacing / 2
    centerx = left + (right - left) / 2
    centery = top + (bottom - top) / 2

    # 1 2 3
    # 4 5 6
    # 7 8 9
    p1 = (left, top)
    p2 = (centerx, top)
    p3 = (right, top)
    p4 = (left, centery)
    p5 = (centerx, centery)
    p6 = (right, centery)
    p7 = (left, bottom)
    p8 = (centerx, bottom)
    p9 = (right, bottom)

    self.outer_loop = [p1, p3, p9, p7]
    self.outer_diamond = [p4, p2, p6, p8]
    self.lower_8 = [p4, p5, p8, p9, p6, p5, p8, p7]
    self.lower_x = [p4, p9, p6, p7]
    self.top = [p1, p2, p3]
    self.bottom = [p7, p8, p9]
    self.star = [p1, p5, p6, p5, p7, p5, p2, p5, p9, p5, p4, p5, p3, p5, p8]

    self.paths = [self.outer_loop, self.outer_diamond, self.lower_8, self.top, self.bottom, self.star]
    self.switch_path()

  def switch_path(self):
    currentx, currenty = self.points[self.loop_pos]
    self.points = random.choice(self.paths)
    new_loop_pos = 0
    for i, pos in enumerate(self.points):
      x, y = pos
      if (x-currentx) ** 2 + (y - currenty) ** 2 > 10000:
        new_loop_pos = i
        break

    self.loop_pos = new_loop_pos

  def choose_arm(self, preferred_rotation):
    if not self.cyborg_arms:
      return None

    next_arm = self.current_arm
    best = 10
    for arm in self.cyborg_arms:
      if arm == self.current_arm:
        continue
      diff = abs(arm.rotation - preferred_rotation) % (2 * math.pi)
      if diff < math.pi / 2:
        return arm
      if diff < best:
        next_arm = arm
    return next_arm
  def animate(self):
    self.head.animate()
    for arm in self.arms:
      if (self.positions[arm] + self.toggle) % 3 == 0:
        arm.animate()
    self.toggle += 1

  def arm_collides(self, rect):
    if not rect.colliderect(self.colliderect):
      return None

    newrect = rect.move(-self.rect.left, -self.rect.top)
    for arm in self.arms:
      # TODO per pixel collision would work better here
      if arm.colliderect.colliderect(newrect):
        return arm

  def collide(self, bullet, explosion_group, animation_group, mixer):
    if self.dead:
      return True
    arm = self.arm_collides(bullet.rect)
    if arm:
      mixer.play_sound('hurt')
      bullet.kill()
      logging.info('HIT ARM')
      arm.health -= bullet.power
      if arm.health < 0:
        mixer.play_sound('explosion')
        if arm == self.current_arm:
          self.current_arm = None
        arm.kill()
        pos = (self.rect.left + arm.rect.centerx, self.rect.top + arm.rect.centery - 2)
        explosion = Explosion(pos)
        explosion_group.add(explosion)
        animation_group.add(explosion)
    elif not len(self.arms) and self.head.rect.colliderect(bullet.rect):
      mixer.play_sound('hurt')
      bullet.kill()
      logging.info('HIT HEAD')
      self.health -= bullet.power
      if self.health < 0:
        mixer.play_sound('explosion')
        self.dead = True
        pos = (self.head.rect.centerx, self.head.rect.centery - 15)
        explosion = Explosion(pos)
        explosion_group.add(explosion)
        animation_group.add(explosion)

  def update(self, ms, player, mixer, enemy_bullets=[], enemy_groups=[]):
    if self.dead:
      raise Nemesis.NOOOOOooo()
    Enemy.update(self, ms, player, enemy_bullets)
    # Redraw arms
    self.image.fill(self.transparent)
    self.head.rect.center = self.rect.center
    self.head.set_sprite(*self.head.frames[self.head.current_frame])
    self.head.image = transform.rotate(self.head.image, -self.rotation * 45)
    width, height = self.head.image.get_size()
    x = self.spacing + self.radius - width / 2
    y = self.spacing + self.radius - height /2
    self.image.blit(self.head.image, (x, y))
    for arm in self.arms:
      i = self.positions[arm]
      theta = (i + self.rotation) * math.pi / 4
      theta_deg = (i + self.rotation) * 45
      arm.rotation = theta
      arm.set_sprite(*arm.frames[arm.current_frame])
      arm_image = arm.image
      rotated = transform.rotate(arm_image, 270-theta_deg)
      arm.image = rotated
      width, height = rotated.get_size()
      x = self.spacing + self.radius + self.radius * math.cos(theta) - width / 2
      y = self.spacing + self.radius + self.radius * math.sin(theta) - height /2
      arm.rect = Rect(x, y, width, height)
      arm.colliderect = arm.rect.inflate(-15, -20)

    self.arms.draw(self.image)
    self.rotation += 0.001 * ms

    self.action_timer -= ms
    if self.moving and not self.firing and self.action_timer < 0:
        direction = math.atan2(player.rect.centery - self.rect.centery, player.rect.centerx - self.rect.centerx)
        direction += math.pi # Move from -pi - pi -> 0 - 2pi
        direction = (direction + math.pi /2) % (math.pi * 2) # change where it is relative to
        next_arm = self.choose_arm(direction)
        self.switch_path()
        if next_arm:
          self.moving = next_arm == self.jellyfish_arm
          self.firing = True
          self.current_arm = next_arm
          self.action_timer = self.fire_max_time
        else:
          # Keep moving
          self.current_arm = None
          self.action_timer = self.move_max_time
          self.firing = False
    elif self.firing:
      if self.action_timer < 0:
        self.switch_path()
        self.action_timer = self.move_max_time
        self.moving = True
        self.firing = False
      elif self.current_arm:
        self.current_arm.fire(ms, player, mixer, self.current_arm.rect.move(self.rect.left, self.rect.top), enemy_bullets, enemy_groups, self.config)

class PlayerBullet(Sprite):
  def __init__(self, pos, speed, *args):
    Sprite.__init__(self, *args)
    self.size = (2,6)
    self.color = (255, 0, 0)
    self.image = Surface(self.size)
    self.image.fill(self.color)
    self.rect = self.image.get_rect()
    self.rect.midtop = pos
    self.speed = speed
    self.power = 50

  def update(self, ms):
    self.rect.top += ms * self.speed

  def change_stats(self, size, color, power):
    self.power = power
    self.size = size
    self.color = color
    self.image = Surface(self.size)
    self.image.fill(self.color)

class NotPlayerBullet(Sprite):
  """
  IOU 500 lines of well designed code - Mat
  """
  def __init__(self, pos, speed, *args):
    Sprite.__init__(self, *args)
    self.size = (2,6)
    self.color = (255, 0, 0)
    self.image = Surface(self.size)
    self.image.fill(self.color)
    self.rect = self.image.get_rect()
    self.rect.midtop = pos
    self.yspeed, self.xspeed = speed

  def update(self, ms):
    self.rect.top += ms * self.yspeed
    self.rect.x += ms * self.xspeed

class SquidBullet(Sprite):
  def __init__(self, pos, speed, *args):
    Sprite.__init__(self, *args)
    self.image = data.image_load('inkball.png')
    self.rect = self.image.get_rect()
    self.rect.midtop = pos
    self.yspeed, self.xspeed = speed
    self.power = 50

  def update(self, ms):
    self.rect.top += ms * self.yspeed
    self.rect.x += ms * self.xspeed

class Squid2Bullet(Sprite):
  def __init__(self, pos, speed, *args):
    Sprite.__init__(self, *args)
    self.image = data.image_load('inkball_large.png')
    self.rect = self.image.get_rect()
    self.rect.midtop = pos
    self.yspeed, self.xspeed = speed
    self.power = 50

  def update(self, ms):
    self.rect.top += ms * self.yspeed
    self.rect.x += ms * self.xspeed
